#region Using Statements
using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
#endregion

namespace Boids {

	public class BoidsComponent : GameComponent {
		readonly Boids boids;
		public BoidsComponent(Game game, Boids boids)
			: base(game) { this.boids = boids; }
		public override void Update(GameTime gameTime) {
			for (int i = 0; i < boids.Count; i++) {
				boids[i].UpdateVelocity();
			}
			for (int i = 0; i < boids.Count; i++) {
				boids[i].Update();
			}
			base.Update(gameTime);
		}
	}

	public enum Mode {
		Wander,
		Separation,
		Alignment,
		Cohesion,
	}

	public class Boids {
		public readonly List<Boid> boids = new List<Boid>();
		public int Count { get { return boids.Count; } }
		public Boid this[int index] { get { return boids[index]; } }
	}
	public class Boid {
		readonly Boids boids;
		Vector3 position;
		Vector3 velocity;
		Quaternion goalRotation = Quaternion.Identity;
		Quaternion rotaiton = Quaternion.Identity;
		Random random = new Random();
		Mode mode;

		public Quaternion Rotation { get { return rotaiton; } }
		public Vector3 Position { get { return position; } }
		public Mode Mode { get { return mode; } }

		public Boid(Boids boids) {
			this.boids = boids;
		}
		public Boid(Boids boids, Vector3 position)
			: this(boids) {
			this.position = position;
		}
		public Vector3? CenterOfFlock {
			get {
				Vector3 center = Vector3.Zero;
				int count = 0;
				foreach (Boid boid in Flock) {
					center += boid.position;
					count++;
				}
				if (count <= 0)
					return null;
				return center /= count;
			}
		}
		public Vector3? VelocityOfFlock {
			get {
				Vector3 velocity = Vector3.Zero;
				int count = 0;
				foreach (Boid boid in Flock) {
					velocity += boid.velocity;
					count++;
				}
				if (count <= 0)
					return null;
				return velocity /= count;
			}
		}

		const float scale = 1.2f;
		const float near = 10.0f * scale;
		const float far = 25.0f * scale;
		const float viewDistance = 35.0f * scale;

		public void UpdateVelocity() {
			Boid closest = Closest;

			Vector3 goalVelocity = Vector3.Zero;

			if (closest != null) {
				float d = Vector3.DistanceSquared(position, closest.position);
				if (d <= near * near) {
					goalVelocity = position - closest.position;
					velocity = goalVelocity * -0.08f;
					mode = Mode.Separation;
				} else if (d >= far * far) {
					Vector3? center = CenterOfFlock;
					if (center != null) {
						goalVelocity = (Vector3)center - position;
						velocity = goalVelocity * 0.2f;
					}
					mode = Mode.Cohesion;
				} else {
					Vector3? flockVelocity = VelocityOfFlock;
					if (flockVelocity != null) {
						goalVelocity = (Vector3)flockVelocity;
						velocity = goalVelocity;
					}
					mode = Mode.Alignment;
				}
			} else {
				if (random.Next(20) == 0) {
					goalVelocity = Forward;
					goalVelocity.X += (float)random.NextDouble() - 0.5f;
					goalVelocity.Y += (float)random.NextDouble() - 0.5f;
					goalVelocity.Z += (float)random.NextDouble() - 0.5f;
					velocity = goalVelocity * 0.3f;//Vector3.Lerp(ref velocity, ref cohesion, 0.3f, out velocity);
				}
				mode = Mode.Wander;
				//velocity = Vector3.Zero;
			}
			//velocity.Y = 0;
			float speedSquared = velocity.LengthSquared();
			if (speedSquared > 0) {
				Vector3 direction = Vector3.Normalize(velocity);
				if (speedSquared > 1) {
					velocity = direction;
				}
				if (Vector3.Dot(direction, Vector3.Up) < 1) {
					Matrix goalRotationMatrix = Matrix.CreateLookAt(Vector3.Zero, direction, Vector3.Up);
					goalRotation = Quaternion.CreateFromRotationMatrix(goalRotationMatrix);
					goalRotation.Normalize();
				}
			}
		}

#if WRAP
		public static float Wrap(float value) {
			if (value < -100)
				return -100;
			else if (value > 100)
				return 100;
			return value;
		}
#else
		public static float Wrap(float value, float minimumValue, float maximumValue) {
			if (value < minimumValue)
				return value + (maximumValue - minimumValue);
			else if (value > maximumValue)
				return value - (maximumValue - minimumValue);
			return value;
		}
#endif

		const float boundary = 100;
		public void Update() {
			position += Forward * 0.5f * velocity.Length();
			//position += velocity;// Forward * 0.1f * velocity.Length();

			position.X = Wrap(position.X, -boundary, +boundary);
			position.Y = Wrap(position.Y, -boundary, +boundary);
			position.Z = Wrap(position.Z, -boundary, +boundary);
			Quaternion.Slerp(ref rotaiton, ref goalRotation, 0.1f, out rotaiton);
		}
		public Boid Closest {
			get {
				float distance = float.MaxValue;
				Boid closest = null;
				foreach (Boid boid in Flock) {
					if (boid == this)
						continue;
					float d = Vector3.DistanceSquared(position, boid.position);
					if (d < distance) {
						d = distance;
						closest = boid;
					}
				}
				return closest;
			}
		}
		public IEnumerable<Boid> Flock {
			get {
				for (int i = 0; i < boids.Count; i++) {
					//if (boids[i] != this)
					if (IsVisible(boids[i]))
						yield return boids[i];
				}
			}
		}
		public bool IsVisible(Boid other) {
			Vector3 direction = other.position - position;
			float speedSquared = direction.LengthSquared();
			if (speedSquared >= viewDistance * viewDistance)
				return false;
			if (speedSquared > 0)
				direction.Normalize();
			return (Vector3.Dot(Forward, direction) >= -0.25f);
			//&& Vector3.Dot(Vector3.Normalize(velocity), Vector3.Normalize(other.velocity)) >= -0.25f;
		}

		public Vector3 Forward {
			get { return Matrix.CreateFromQuaternion(rotaiton).Backward; }
		}
	}

	public class BoidsVisual : DrawableGameComponent {
		readonly Boids boids;
		Model model;
		readonly ContentManager content;
		Matrix[] transforms = new Matrix[32];
		Matrix projection = Matrix.Identity;
		Matrix view = Matrix.Identity;

		public static readonly Vector3[] colors = {
			new Vector3(0.35f, 0.35f, 0.35f),	// white
			new Vector3(0.35f, 0.10f, 0.00f),	// red
			new Vector3(0.00f, 0.35f, 0.20f),	// green
			new Vector3(0.10f, 0.00f, 0.35f),	// blue
		};

		public BoidsVisual(Game game, ContentManager content, Boids boids)
			: base(game) {
			this.boids = boids;
			this.content = content;
		}
		protected override void LoadGraphicsContent(bool loadAllContent) {
			if (loadAllContent) {
				model = content.Load<Model>("content/model/boid");
			}
			base.LoadGraphicsContent(loadAllContent);
		}
		public override void Draw(GameTime gameTime) {

			float aspectRatio = (float)GraphicsDevice.Viewport.Width / GraphicsDevice.Viewport.Height;
			Matrix.CreatePerspectiveFieldOfView(MathHelper.ToRadians(45.0f),
			   aspectRatio, 0.01f, 5.00f, out projection);
			view = Matrix.CreateLookAt(new Vector3(0.00f, -0.50f, 1.75f), Vector3.Zero, Vector3.Up);

			for (int i = 0; i < boids.Count; i++) {
				DrawModel(boids[i].Rotation, boids[i].Position, colors[(int)boids[i].Mode]);
			}
			base.Draw(gameTime);
		}

		static readonly Vector3 fog = new Vector3(0.5f, 0.4f, 0.7f);
		static readonly Vector3 ambient = new Vector3(0.5f, 0.4f, 0.7f);
		static readonly Vector3 specular = new Vector3(1.2f, 1.1f, 1.0f);
		void DrawModel(Quaternion rotation, Vector3 position, Vector3 color) {
			Matrix local = Matrix.CreateFromQuaternion(rotation);

			if (transforms.Length < model.Bones.Count) {
				Array.Resize<Matrix>(ref transforms, model.Bones.Count);
			}
			model.CopyAbsoluteBoneTransformsTo(transforms);
			foreach (ModelMesh mesh in model.Meshes) {
				foreach (BasicEffect effect in mesh.Effects) {
					effect.PreferPerPixelLighting = true;
					effect.EnableDefaultLighting();
					effect.FogEnabled = true;
					effect.FogEnd = 4.0f;
					effect.FogStart = 2.0f;
					effect.FogColor = fog;
					effect.SpecularPower = 15;
					effect.SpecularColor = specular;
					effect.AmbientLightColor = ambient;
					effect.DiffuseColor = color;
					effect.View = view;
					effect.Projection = projection;
					effect.World = local * transforms[mesh.ParentBone.Index]
						* Matrix.CreateTranslation(position) * Matrix.CreateScale(0.01f);
				}
				mesh.Draw();
			}
		}
	}

}
